import ast
import inspect

from relsa_serializer.tools_modules.work_with_function_module.ast_to_dict_module import ast_to_dict


def serialize_to_dict(obj: object, globals_from_main: dict) -> dict:
    """
    Перевод объекта в словарь для сериализации
    """
    result = dict()
    globals().update(globals_from_main)
    if isinstance(obj, int) or isinstance(obj, float) or isinstance(obj, str) or isinstance(obj, bool):
        result['type'] = type(obj).__name__
        result['value'] = obj

    elif inspect.isroutine(obj):
        funct = inspect.getsource(obj)
        if funct[:4] == '    ':
            funct = funct[4:]
        ast_tree = ast.parse(funct)
        ser_str = ast_to_dict(ast_tree)
        result[str(type(obj))] = ser_str

    elif inspect.isclass(obj):
        result['type'] = 'class'
        result['class_name'] = obj.__name__
        class_members = inspect.getmembers(obj)
        class_initializers = [i for i in class_members if i[0] == '__init__']
        class_members = [i for i in class_members if i[0][0:2] != '__']
        class_members = class_members + class_initializers

        for class_member in class_members:
            result[class_member[0]] = serialize_to_dict(class_member[1], globals_from_main)

    elif isinstance(obj, dict):
        for item in obj:
            if isinstance(item, int) or isinstance(item, float) or isinstance(item, str) or isinstance(item, bool):
                wrk_dict = dict()
                wrk_dict['type'] = type(obj[item]).__name__
                wrk_dict['value'] = obj[item]
                result[item] = wrk_dict

    elif isinstance(obj, list) or isinstance(obj, tuple):
        wrk_dict = dict()

        if isinstance(obj, list):
            wrk_dict['type'] = 'list'
        else:
            wrk_dict['type'] = 'tuple'

        for item in obj:
            if isinstance(item, int) or isinstance(item, float) or isinstance(item, str) or isinstance(item, bool):
                wrk_tmp = dict()
                wrk_tmp['type'] = type(item).__name__
                wrk_tmp['value'] = item
                wrk_dict[item] = wrk_tmp
        result = wrk_dict

    else:
        result['type'] = 'instance'
        result['class'] = serialize_to_dict(type(obj), globals_from_main)
        obj_members = inspect.getmembers(obj)
        obj_members = [i for i in obj_members if i[0][0:2] != '__']
        for class_member in obj_members:
            result[class_member[0]] = serialize_to_dict(class_member[1], globals_from_main)

    return result



from abc import abstractmethod


class AbstractSerializer:
    @abstractmethod
    def dump(self, target_item, file_path: str, globals_from_main: dict) -> None:
        pass

    @abstractmethod
    def dumps(self, target_item, globals_from_main: dict) -> str:
        pass

    @abstractmethod
    def load(self, file_path: str, globals_from_main) -> all:
        pass

    @abstractmethod
    def loads(self, target_str, globals_from_main) -> all:
        pass



list_of_asts = ["Module", "Interactive", "Expression", "FunctionType", "FunctionDef", "AsyncFunctionDef",
                "ClassDef", "Return", "Delete", "Assign", "AugAssign", "AnnAssign", "For", "AsyncFor", "While",
                "If", "With", "AsyncWith", "Match", "Raise", "Try", "Assert", "Import", "ImportFrom", "Global",
                "Nonlocal", "Expr", "Pass", "Break", "Continue", "BoolOp", "NamedExpr", "BinOp", "UnaryOp",
                "Lambda", "IfExp", "Dict", "Set", "ListComp", "SetComp", "DictComp", "GeneratorExp", "Await",
                "Yield", "YieldFrom", "Compare", "Call", "FormattedValue", "JoinedStr", "Constant", "Attribute",
                "Subscript", "Starred", "Name", "List", "Tuple", "Slice", "Load", "Store", "Del", "And", "Or",
                "Add", "Sub", "Mult", "MatMult", "Div", "Mod", "Pow", "LShift", "RShift", "BitOr", "BitXor",
                "BitAnd", "FloorDiv", "Invert", "Not", "UAdd", "USub", "Eq", "NotEq", "Lt", "LtE", "Gt", "GtE",
                "Is", "IsNot", "In", "NotIn", "ExceptHandler", "MatchValue", "MatchSingleton", "MatchSequence",
                "MatchMapping", "MatchClass", "MatchStar", "MatchAs", "MatchOr", "TypeIgnore", "arguments", "arg"]


def unparse_ast_str(unparse_ast: str):
    """
    Перевод ast-словаря в форму, пригодную для инициализации заполнения узла
    """
    result_str = unparse_ast.replace(":", "=")
    result_str = result_str[2:]
    result_str = result_str[:len(result_str) - 1]

    for i in range(0, len(result_str)):
        if i == len(result_str):
            break

        if result_str[i] == "'" and result_str[i + 1] == "=":
            result_str = result_str[:i] + result_str[i + 1:]
            for j in range(i, 0, -1):
                if result_str[j] == "'":
                    result_str = result_str[:j] + result_str[j + 1:]
                    break

    for i in range(0, len(result_str)):
        if i == len(result_str):
            break

        flag = 0
        if result_str[i] == "=":
            for j in range(i - 1, -1, -1):
                d = result_str[j]
                if not result_str[j].isalpha():
                    break
                if not result_str[j].islower():
                    flag = 1
                    break

        if flag == 1:
            result_str = result_str[:i] + result_str[i + 1:]

    result_str = result_str.replace("={}", "=[]")
    wrk_list_of_asts = set(list_of_asts)
    for i in wrk_list_of_asts:
        str_1 = i + "{"
        str_2 = i + "("
        result_str = result_str.replace(str_1, str_2)

        str_3 = i + "={"
        result_str = result_str.replace(str_3, str_2)

    result_str = close_all_brekets(result_str)
    result_str = del_figure_braket(result_str)
    result_str = set_dot_ast(result_str)

    result_str = result_str.replace("}", "]")
    result_str = result_str.replace("{", "[")
    return result_str


def set_dot_ast(unparse_ast: str):
    """
    Добавляет 'ast.' перед узлами дерева
    """
    wrk_list_of_asts = set(list_of_asts)
    unparse_ast = "ast." + unparse_ast
    for i in wrk_list_of_asts:
        unparse_ast = unparse_ast.replace(f"{i}(", f"ast.{i}(")

    return unparse_ast[4:]


def close_all_brekets(inpt_str: str):
    """
    Закрывает скобки
    """
    start_index = 0
    for _ in range(0, inpt_str.count('(')):
        current_index = inpt_str.find('(', start_index)
        wrk_ind = current_index
        count = 0
        while True:
            if inpt_str[wrk_ind] == '(' or inpt_str[wrk_ind] == '{':
                count += 1
            elif inpt_str[wrk_ind] == ')' or inpt_str[wrk_ind] == '}':
                count -= 1

            if count == 0:
                inpt_str = inpt_str[:wrk_ind] + ')' + inpt_str[wrk_ind + 1:]
                break
            wrk_ind += 1

        start_index = current_index + 1

    return inpt_str


def del_figure_braket(inpt_str: str):
    """
    Удаление остатков фигурных скобок из строки
    """
    start_index = 0
    for _ in range(0, inpt_str.count('{')):
        current_index = inpt_str.find('{', start_index)
        start_index = current_index + 1
        if inpt_str[current_index - 2] == 'y' and inpt_str[current_index - 3] == 'd' and \
                inpt_str[current_index - 4] == 'o' and inpt_str[current_index - 5] == 'b':
            continue
        elif inpt_str[current_index - 2] == 's' and inpt_str[current_index - 3] == 'g' and \
                inpt_str[current_index - 4] == 'r' and inpt_str[current_index - 5] == 'a' and \
                inpt_str[current_index + 4] != 'u':
            continue
        elif inpt_str[current_index - 2] == 's' and inpt_str[current_index - 3] == 't' and \
                inpt_str[current_index - 4] == 'e' and inpt_str[current_index - 5] == 'g' and \
                inpt_str[current_index - 6] == 'r' and inpt_str[current_index - 7] == 'a' and \
                inpt_str[current_index - 8] == 't':
            continue
        else:
            wrk_index = current_index
            while True:
                if inpt_str[wrk_index] == '(':
                    break
                wrk_index += 1

            count = 0
            while True:
                if inpt_str[wrk_index] == '(':
                    count += 1
                elif inpt_str[wrk_index] == ')':
                    count -= 1

                if count == 0:
                    if not inpt_str[wrk_index + 1] == ',':
                        inpt_str = inpt_str[:current_index] + inpt_str[current_index + 1:]
                        inpt_str = inpt_str[:wrk_index] + inpt_str[wrk_index + 1:]
                    break
                wrk_index += 1

    return inpt_str



import ast

import toml
import yaml


def transfer_to_another_format(input_file: str, input_file_format: str, output_file_name: str,
                               output_file_format: str) -> bool:
    """
    Конвертирует один формат сериализации в другой
    """
    if output_file_format.lower() == 'yaml':
        if input_file_format.lower() == 'toml' and 'toml' in input_file:
            restored_class_toml = toml.load(input_file)
            with open(f"{output_file_name}.yaml", 'w+') as out_file:
                yaml.dump(restored_class_toml, out_file)
            return True

        elif input_file_format.lower() == 'json' and 'json' in input_file:
            with open(input_file, 'r+') as file:
                restored_class_json = ast.literal_eval(file.read())
            with open(f"{output_file_name}.yaml", 'w+') as out_file:
                yaml.dump(restored_class_json, out_file)
            return True

        elif input_file_format.lower() == 'yaml' and 'yaml' in input_file:
            return True

        return False

    elif output_file_format.lower() == 'toml':
        if input_file_format.lower() == 'yaml' and 'yaml' in input_file:
            with open(input_file, 'r+') as in_file:
                restored_class_yaml = yaml.safe_load(in_file.read())
            with open(f"{output_file_name}.toml", 'w+') as out_file:
                toml.dump(restored_class_yaml, out_file)
            return True

        elif input_file_format.lower() == 'json' and 'json' in input_file:
            with open(input_file, 'r+') as file:
                restored_class_json = ast.literal_eval(file.read())
            with open(f"{output_file_name}.toml", 'w+') as out_file:
                toml.dump(restored_class_json, out_file)
            return True

        elif input_file_format.lower() == 'toml' and 'toml' in input_file:
            return True

        return False

    elif output_file_format.lower() == 'json':
        if input_file_format.lower() == 'toml' and 'toml' in input_file:
            restored_class_toml = toml.load(input_file)
            with open(f"{output_file_name}.json", 'w+') as file:
                file.write(str(restored_class_toml))
            return True

        elif input_file_format.lower() == 'yaml' and 'yaml' in input_file:
            with open(input_file, 'r+') as in_file:
                restored_class_yaml = yaml.safe_load(in_file.read())
            with open(f"{output_file_name}.json", 'w+') as file:
                file.write(str(restored_class_yaml))
            return True

        elif input_file_format.lower() == 'json' and 'json' in input_file:
            return True

        return False



import ast
import types


def get_function_name(unparse_str: str) -> str:
    """
    Получение названия функции
    """
    result = ""
    for i in range(4, len(unparse_str)):
        if unparse_str[i] == '(':
            return result
        result += unparse_str[i]


def unparsed_funct(inpt_ast: str):
    """
    Перевод разименованной ast-строки в дерево
    """
    ss = eval(inpt_ast)
    ast.fix_missing_locations(ss)
    unparse_ast = ast.unparse(ss)
    return unparse_ast


def return_compiled_function(new_str: str, globals_main):
    """
    Возвращает рабочую функцию
    """
    unparse = unparsed_funct(new_str)
    ast_node = ast.parse(unparse)
    code = compile(ast_node, '<ast>', mode="exec")
    f_code = next(c for c in code.co_consts if isinstance(c, types.CodeType))
    compiled_function = types.FunctionType(f_code, globals_main)
    return compiled_function



from relsa_serializer.serialize_modules.json_serialize_module import JsonSerializer
from relsa_serializer.serialize_modules.toml_serialize_module import TomlSerializer
from relsa_serializer.serialize_modules.yaml_serialize_module import YamlSerializer


class SerializeFactory:

    @staticmethod
    def create_serializer(name_of_serializer: str) -> object:
        """
        Возвращает сериализатор по имени формата
        """
        lw_name = name_of_serializer.lower()
        if lw_name == "json":
            return JsonSerializer()
        elif lw_name == "toml":
            return TomlSerializer()
        elif lw_name == "yaml":
            return YamlSerializer()



import yaml

from relsa_serializer.serialize_modules.abstractserializer import AbstractSerializer
from relsa_serializer.tools_modules.convert_to_universal_format import serialize_to_dict
from relsa_serializer.tools_modules.return_to_normal_form import deserialize_from_dict


class YamlSerializer(AbstractSerializer):
    def dump(self, target_item, file_path: str, globals_from_main: dict) -> None:
        """
        Сериализация объекта в файл
        """
        with open(file_path, 'w+') as file:
            file.write(yaml.dump(serialize_to_dict(target_item, globals_from_main)))
            file.close()

    def dumps(self, target_item, globals_from_main: dict) -> str:
        """
        Сериализация объекта в строку
        """
        return yaml.safe_dump(serialize_to_dict(target_item, globals_from_main))

    def load(self, file_path: str, globals_from_main) -> all:
        """
        Возвращает объект из файла
        """
        with open(file_path, 'r+') as file:
            wrk_str = file.read()
            wrk_str = yaml.safe_load(wrk_str)
            file.close()
        return deserialize_from_dict(wrk_str, globals_from_main)

    def loads(self, target_str, globals_from_main) -> all:
        """
        Возвращает объект из строки
        """
        wrk_str = yaml.safe_load(target_str)
        return deserialize_from_dict(wrk_str, globals_from_main)



from setuptools import setup

long_description = '''Simple Python module to serialize objects to json, toml or yaml'''

setup(
    name="relsa_serializer",
    packages=["relsa_serializer", "relsa_serializer.serialize_modules", "relsa_serializer.tools_modules",
              "relsa_serializer.tools_modules.work_with_function_module"],
    version="1.0.0",
    author="_relsa",
)



from relsa_serializer.tools_modules.work_with_function_module.dic_to_ast_module import unparse_ast_str
from relsa_serializer.tools_modules.work_with_function_module.unserialize_funct_module import return_compiled_function
from relsa_serializer.tools_modules.work_with_function_module.unserialize_funct_module import unparsed_funct


def deserialize_from_dict(incoming_dict: dict, globals_from_main) -> object:
    """
    Возвращает объект из универсального словаря
    """
    if 'type' in incoming_dict and incoming_dict['type'] == 'class':
        return class_deserialize(incoming_dict, globals_from_main)

    elif 'type' in incoming_dict and incoming_dict['type'] == 'instance':
        return instance_deserialize(incoming_dict, globals_from_main)

    elif "<class 'function'>" in incoming_dict:
        return function_deserialize(incoming_dict, globals_from_main)

    if 'type' in incoming_dict and incoming_dict['type'] == 'list':
        return list_deserialize(incoming_dict)

    elif 'type' in incoming_dict and incoming_dict['type'] == 'tuple':
        return tuple_deserialize(incoming_dict)

    else:
        else_dict = dict()

        for item in incoming_dict.items():
            if type(item[1]) == dict and 'value' in item[1]:
                else_dict[item[0]] = item[1]['value']

        return else_dict


def class_deserialize(incoming_dict: dict, globals_from_main):
    dict_as_obj = dict()
    for item in incoming_dict.items():
        if type(item[1]) == dict and "<class 'function'>" in item[1]:
            unparsed_funct(unparse_ast_str(item[1]["<class 'function'>"]))
            unprs_fnct = return_compiled_function(unparse_ast_str(item[1]["<class 'function'>"]), globals_from_main)
            dict_as_obj[item[0]] = unprs_fnct

    constructed_class = type(f'{incoming_dict["class_name"]}', (object,), dict_as_obj)

    return constructed_class


def instance_deserialize(incoming_dict: dict, globals_from_main):
    class_deserialized = deserialize_from_dict(incoming_dict['class'], globals_from_main)

    params_dict = dict()
    for item in incoming_dict.items():
        if type(item[1]) == dict and 'value' in item[1]:
            params_dict[item[0]] = item[1]['value']
        elif type(item[1]) == dict and 'type' in item[1] and item[1]['type'] == 'instance' and len(
                params_dict) != 0:
            if class_deserialized:
                for item_in_included in item[1].items():
                    if type(item_in_included) == tuple and 'value' in item_in_included[1]:
                        params_dict[item_in_included[0]] = item_in_included[1]['value']

    constructed_instance = class_deserialized(**params_dict)
    return constructed_instance


def function_deserialize(incoming_dict: dict, globals_from_main):
    unparsed_funct(unparse_ast_str(incoming_dict["<class 'function'>"]))
    unprs_fnct = return_compiled_function(unparse_ast_str(incoming_dict["<class 'function'>"]), globals_from_main)
    return unprs_fnct


def list_deserialize(incoming_dict: dict):
    out_list = list()

    for item in incoming_dict.items():
        if 'value' in item[1]:
            out_list.append(type(item[1]['type'])(item[1]['value']))

    return out_list


def tuple_deserialize(incoming_dict: dict):
    out_tuple = tuple()

    for item in incoming_dict.items():
        if 'value' in item[1]:
            out_tuple = list(out_tuple)
            out_tuple.append(type(item[1]['type'])(item[1]['value']))
            out_tuple = tuple(out_tuple)

    return out_tuple



import ast


def ast_to_dict(ast_node) -> str:
    """
    Переводит ast узел в словарь
    """
    wrk_str = ast.dump(ast_node)
    wrk_str = wrk_str.replace('=', ':')
    wrk_str = wrk_str.replace('(', ":{")
    wrk_str = wrk_str.replace(')', '}')
    wrk_str = wrk_str.replace('[', '{')
    wrk_str = wrk_str.replace(']', '}')
    wrk_str = "{" + wrk_str + "}"

    start_index = 0
    while True:
        current_index = wrk_str.find(':', start_index)
        if wrk_str[current_index + 1].isalpha():
            wrk_str = wrk_str[:current_index + 1] + '[' + wrk_str[current_index + 1:]

        if current_index == -1:
            break
        else:
            start_index = current_index + 1

    wrk_str = close_all_brekets(wrk_str)
    wrk_str = wrk_str.replace('[', '{')
    wrk_str = wrk_str.replace(']', '}')

    wrk_str = add_quotes(wrk_str)
    return wrk_str


def close_all_brekets(inpt_str: str) -> str:
    """
    Закрывает открытые скобки (для поиска списков)
    """
    start_index = 0
    for _ in range(0, inpt_str.count('[')):
        current_index = inpt_str.find('[', start_index)

        wrk_ind = current_index
        while True:
            if inpt_str[wrk_ind] == '{':
                break
            wrk_ind += 1

        count = 0
        while True:
            if inpt_str[wrk_ind] == '{':
                count += 1
            elif inpt_str[wrk_ind] == '}':
                count -= 1

            if count == 0:
                inpt_str = inpt_str[:wrk_ind + 1] + ']' + inpt_str[wrk_ind + 1:]
                break
            wrk_ind += 1

        start_index = current_index + 1

    return inpt_str


def add_quotes(inpt_str: str) -> str:
    """
    Названия ast полей берутся в кавычки
    """
    i = 0
    while True:
        if i == len(inpt_str):
            break
        if inpt_str[i].isalpha() and inpt_str[i - 1] != '\'':
            inpt_str = inpt_str[:i] + '\'' + inpt_str[i:]
            i += 1
            while inpt_str[i].isalpha() or inpt_str[i] == '_':
                i += 1
            inpt_str = inpt_str[:i] + '\'' + inpt_str[i:]
        elif inpt_str[i] != '\'':
            while inpt_str[i].isalpha() or inpt_str[i] == '_':
                i += 1
        i += 1
    return inpt_str



from relsa_serializer.serialize_modules.serialize_factory import SerializeFactory
from relsa_serializer.serialize_modules.formats_converter import transfer_to_another_format

__author__ = '_relsa'
__version__ = '1.0.0'
